#include <unistd.h>
#include "ucspissl.h"
#include "fmt.h"
#include "stralloc.h"
#include "str.h"
#include "byte.h"
#include "case.h"
#include "dns.h"
#include "constmap.h"
#include "tls_remote.h"
#include "tls_errors.h"

/** @file  tls_remote.c -- TLS client functions
    @brief connection functions: tls_conn, tls_exit;
           verification functions: tls_certkey, tls_checkpeer, tls_fingerprint, tlsa_check;
           tls_destination, tls_domaincert
           dummy functions: tls_crlcheck

           tls_checkpeer:   r = 0 -> ADH, r = 1 -> wildcard DN, r = 2 -> DN, r = 3 -> CA; r < 0 -> error
           tls_fingerprint: r = 0 -> failed, r = 1 -> ok; r < 0 -> error
           tlsa_check:      r = 0 -> nothing, r = 1 -> sha256, r = 2 -> sha512; r < 0 -> error
*/

int tls_certkey(SSL_CTX *ctx,const char *cert,const char *key,char *ppwd) 
{
  if (!cert) return 0;

  if (SSL_CTX_use_certificate_chain_file(ctx,cert) != 1)
    return -1;

  if (!key) key = cert;

  if (ppwd) SSL_CTX_set_default_passwd_cb_userdata(ctx,ppwd);

  if (SSL_CTX_use_PrivateKey_file(ctx,key,SSL_FILETYPE_PEM) != 1)
    return -2;

  if (SSL_CTX_check_private_key(ctx) != 1)
    return -3;

  return 0;
}

int tls_conn(SSL *ssl,int smtpfd)
{
  SSL_set_options(ssl,SSL_OP_NO_SSLv2);
  SSL_set_options(ssl,SSL_OP_NO_SSLv3);
  return SSL_set_fd(ssl,smtpfd);
}

int tls_checkpeer(SSL *ssl,const char *host,int hlen,const int flag,const int verify)
{
  X509 *cert;
  STACK_OF(GENERAL_NAME) *extensions;
  const GENERAL_NAME *ext;
  char buf[SSL_NAME_LEN];
  char *dnsname = 0;
  int dname = 0;
  int num;
  int len;
  int fflag; 
  int i;
  int rc = 0;

  fflag = flag;
  if (flag > 20) fflag = flag - 20;
  if (flag > 10) fflag = flag - 10;

  if (fflag == 1 || fflag == 3) return 0;       /* ADH - no Cert required */

  cert = SSL_get_peer_certificate(ssl);
  if (!cert) return -1;

  if (host && fflag > 4) {
    extensions = (GENERAL_NAME *)X509_get_ext_d2i(cert,NID_subject_alt_name,0,0);
    num = sk_GENERAL_NAME_num(extensions);	/* num = 0, if no SAN extensions */

    for (i = 0; i < num; ++i) {
      ext = sk_GENERAL_NAME_value(extensions,i);
      if (ext->type == GEN_DNS) {
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) // 0xmnnffppsL
        if (ASN1_STRING_type(ext->d.ia5) != V_ASN1_IA5STRING) continue;
        dnsname = (char *)ASN1_STRING_data(ext->d.ia5);
 #else
        if (OBJ_sn2nid((const char*)ext->d.ia5) != V_ASN1_IA5STRING) continue;
        dnsname = (char *)ASN1_STRING_get0_data(ext->d.ia5);
 #endif
        len = ASN1_STRING_length(ext->d.ia5);
        dname = 1;
      }
    }

    if (!dname) {
      X509_NAME_get_text_by_NID(X509_get_subject_name(cert),NID_commonName,buf,sizeof(buf));
      buf[SSL_NAME_LEN - 1] = 0;
      dnsname = buf;
      len = SSL_NAME_LEN - 1;
    }

    X509_free(cert); 

    switch (fflag) {
      case 5: if (dnsname[0] == '*' && dnsname[1] == '.') 
                if (!case_diffrs(host,dnsname + 1)) return -3;
      case 6: if (str_diffn(host,dnsname,hlen)) return -3;
      default: rc = 1;
    }
  }

  if (fflag > 3 && verify > 0) { 	/* Root CA must be availble */
    if (SSL_get_verify_result(ssl) == X509_V_OK) return -2;
    else rc += 2;
  }

  return rc;
}

int tls_checkcrl(SSL *ssl)   // not implemented yet
{

  return 0;
}

int dig_ascii(char *digascii,const char *digest,const int len)
{
  static const char hextab[] = "0123456789abcdef";
  int j;

  for (j = 0; j < len; j++) {
    digascii[2 * j] = hextab[digest[j] >> 4];
    digascii[2 * j + 1] = hextab[digest[j] & 0x0f];
  }   
  digascii[2 * len] = '\0';

  return (2*j);  // 2*len
}

/* Return codes: -4: no X.509 cert (fatal), -3: matching error (deferred), 
                 -2: unsupported type, -1: weird TLSA record
                  0: No X.509 cert, 1: Type 1 match, 2: Type 2 match */

int tlsa_check(SSL *ssl,const char *host,int hlen,const unsigned long p)
{
  X509 *cert;
  const EVP_MD *methodsha256 = EVP_sha256();
  const EVP_MD *methodsha512 = EVP_sha512();
  stralloc out = {0};
  stralloc sa = {0};
  unsigned char digest[EVP_MAX_MD_SIZE];
  unsigned char fp_tlsa[257];
  unsigned char fp_cert[257];
  unsigned int len;
  int i, rc;
  char port[FMT_ULONG];
  uint16 type;
  uint16 selector;

  cert = SSL_get_peer_certificate(ssl);
  if (!cert) return 0;

// construct TLSA FQDN -- simple procedure; not taking Cert into consideration 

  if (!stralloc_copyb(&sa,"_",1)) temp_nomem();
  port[fmt_ulong(port,p)] = 0;
  if (!stralloc_cats(&sa,port)) temp_nomem();
  if (!stralloc_cats(&sa,"._tcp.")) temp_nomem();
  if (!stralloc_catb(&sa,host,hlen)) temp_nomem();

  if (dns_tlsa(&out,&sa) <= 0) return 0;
  if (out.len < 5) return -1;

  for (i = 0; i <= out.len; i++) {
    selector = (unsigned char) out.s[i + 1]; // Selector: 0 = Cert, 1 = SPKI
    type = (unsigned char) out.s[i + 2];     // Type: 1 = SHA256, 2 = SHA512, Type: 3 = Cert
    i += 3;
    rc = 0;

    if (type == 1) {
      if (selector == 0) rc = X509_digest(cert,methodsha256,digest,&len);
      if (selector == 1) rc = X509_pubkey_digest(cert,methodsha256,digest,&len);
    } else if (type == 2) {
      if (selector == 0) rc = X509_digest(cert,methodsha512,digest,&len);
      if (selector == 1) rc = X509_pubkey_digest(cert,methodsha512,digest,&len);
    } else { rc = -2; goto DONE; }

    if (rc) {    
      dig_ascii(fp_tlsa,out.s + i,len);
      dig_ascii(fp_cert,digest,len); 
      if (!str_diffn(fp_tlsa,fp_cert,2 * len)) { rc = type; goto DONE; } 
    } else { rc = -3; }                      // no matching fingerprint
 
    i += len;
  }
  if (!rc) rc = -4; 

DONE:
  X509_free(cert); 
  return rc;
}

int tls_fingerprint(SSL *ssl,const char *fingerprint,int dlen)
{
  X509 *cert;
  const EVP_MD *methodsha1 = EVP_sha1();
  const EVP_MD *methodsha224 = EVP_sha224();
  const EVP_MD *methodsha256 = EVP_sha256();
  const EVP_MD *methodsha512 = EVP_sha512();
  unsigned char digest[EVP_MAX_MD_SIZE];
  unsigned char digascii[257];
  unsigned int len;

  cert = SSL_get_peer_certificate(ssl);
  if (!cert) return -1;

  switch (dlen) {			/* fetch digest from cert; len = bitlength/8 */
    case  40: if (!X509_digest(cert,methodsha1,digest,&len)) 
                { X509_free(cert); return -2; }; break;
    case  56: if (!X509_digest(cert,methodsha224,digest,&len)) 
                { X509_free(cert); return -2; }; break;
    case  64: if (!X509_digest(cert,methodsha256,digest,&len))
                { X509_free(cert); return -2; }; break;
    case 128: if (!X509_digest(cert,methodsha512,digest,&len))
                { X509_free(cert); return -2; }; break;
    default:  X509_free(cert); return -3; 
  }

  X509_free(cert);  
  len = dig_ascii(digascii,digest,len);
  if (!str_diffn(digascii,fingerprint,len)) return 1;

  return 0;
}

int tls_exit(SSL *ssl)
{
  if (SSL_shutdown(ssl) == 0)
    SSL_shutdown(ssl);

  return 0;
}

/** @brief tls_destination
    @param  pointer to remote fqdn, length of fqdn
                               
                               Certificate       Fallthru

    @return values: | ADH | Cert *DN FQDN Hash | noDANE TLS
          ----------+-----+--------------------+-----------
       optional TLS |  1  |  2    -    -    -  |   8     9
      mandatory TLS |  3  |  4    5    6    7  |   8

             no TLS   -1 
  */

int tls_destination(const stralloc hostname)
{
  int i;
  stralloc tlsdest = {0};
  int len = hostname.len;

// Host rules

  if (!stralloc_copys(&tlsdest,"!")) temp_nomem();
  if (!stralloc_cat(&tlsdest,&hostname)) temp_nomem();
  if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return -1;

  if (!stralloc_copys(&tlsdest,"?")) temp_nomem();
  if (!stralloc_cat(&tlsdest,&hostname)) temp_nomem();
  if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 9;

  if (!stralloc_copys(&tlsdest,"/")) temp_nomem();
  if (!stralloc_cat(&tlsdest,&hostname)) temp_nomem();
  if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 8;

  if (!stralloc_copys(&tlsdest,"%")) temp_nomem(); // CERT + hash 
  if (!stralloc_cat(&tlsdest,&hostname)) temp_nomem();
  if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 7;

  if (!stralloc_copys(&tlsdest,"=")) temp_nomem();  // CERT + FQDN
  if (!stralloc_cat(&tlsdest,&hostname)) temp_nomem();
  if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 6;
  
// Domain rules

  for (i = 0; i <= len; ++i)  // TLS fallthru
    if ((i == 0) || (hostname.s[i] == '.')) {
      if (!stralloc_copys(&tlsdest,"?")) temp_nomem();
      if (!stralloc_cats(&tlsdest,hostname.s + i)) temp_nomem();
      if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 9;
    }

  for (i = 0; i <= len; ++i)  // no DANE 
    if ((i == 0) || (hostname.s[i] == '.')) {
      if (!stralloc_copys(&tlsdest,"/")) temp_nomem();
      if (!stralloc_cats(&tlsdest,hostname.s + i)) temp_nomem();
      if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 8;
    }

  for (i = 0; i <= len; ++i)  // CERT + *DN
    if ((i == 0) || (hostname.s[i] == '.')) {
      if (!stralloc_copys(&tlsdest,"~")) temp_nomem();
      if (!stralloc_cats(&tlsdest,hostname.s + i)) temp_nomem();
      if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 5;
    }

  for (i = 0; i <= len; ++i) // CERT - generic
    if ((i == 0) || (hostname.s[i] == '.')) {
      if (!stralloc_copys(&tlsdest,"")) temp_nomem();
      if (!stralloc_cats(&tlsdest,hostname.s + i)) temp_nomem();
      if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 4;
    }

  for (i = 0; i <= len; ++i)  // ADH per host/domain
    if ((i == 0) || (hostname.s[i] == '.')) {
      if (!stralloc_copys(&tlsdest,"-")) temp_nomem();
      if (!stralloc_cats(&tlsdest,hostname.s + i)) temp_nomem();
      if ((tlsdestinfo = constmap(&maptlsdestinations,tlsdest.s,tlsdest.len))) return 3;
    }

// General rules (mandatory TLS)

  tlsdestinfo = 0;
  if (constmap(&maptlsdestinations,"/*",2)) return 8; // no DANE
  if (constmap(&maptlsdestinations,"=*",2)) return 6; // CERT + FQDN
  if (constmap(&maptlsdestinations,"~*",2)) return 5; // CERT + Wild
  if (constmap(&maptlsdestinations,"+*",2)) return 4;
  if (constmap(&maptlsdestinations,"-*",2)) return 3;

// Fall thru rules (optional TLS)

  if (constmap(&maptlsdestinations,"?",1)) return 9; // fallback to no TLS
  if (constmap(&maptlsdestinations,"/",1)) return 8; // no DANE
  if (constmap(&maptlsdestinations,"*",1)) return 2; // CERT
  if (constmap(&maptlsdestinations,"-",1)) return 1; // ADH

  return 0;
}

int tls_domaincerts(const stralloc domainname)
{
  int i;
  tlsdomaininfo = 0;	// extern

/* Our Certs - per domain */

  if (domainname.len)
    for (i = 0; i <= domainname.len; ++i)
      if ((i == 0) || (domainname.s[i] == '.'))
        if ((tlsdomaininfo = constmap(&mapdomaincerts,domainname.s,domainname.len - i))) return 2;

/* Standard Cert (if any) */
 
  if ((tlsdomaininfo = constmap(&mapdomaincerts,"*",1))) return 1;

  return 0;
}
